import torch
import torch.nn as nn
from einops.einops import rearrange
import math
import copy
from timm.layers import DropPath, trunc_normal_
from .quadtree_attention import QuadtreeAttention
from .common import torch_init_model
from .resnet_fpn import ResNetFPN_8_2


class DWConv(nn.Module):
    def __init__(self, dim=768):
        super(DWConv, self).__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, bias=True, groups=dim)

    def forward(self, x, H, W):
        B, N, C = x.shape
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)
        x = x.flatten(2).transpose(1, 2)

        return x


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)
        self.relu = nn.ReLU(inplace=True)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = self.fc1(x)
        x = self.relu(x)

        x = self.dwconv(x, H, W)

        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)

        return x


class QuadtreeBlock(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        topks,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
        scale=1,
        attn_type="B",
    ):
        super().__init__()

        self.norm1 = norm_layer(dim)

        self.attn = QuadtreeAttention(
            dim,
            num_heads=num_heads,
            topks=topks,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
            scale=scale,
            attn_type=attn_type,
        )

        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here

        self.norm2 = norm_layer(dim)

        mlp_hidden_dim = int(dim * mlp_ratio)

        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if hasattr(m, "init"):
            return
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, target, H, W):
        x = x + self.drop_path(self.attn(self.norm1(x), self.norm1(target), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))

        return x


class LocalFeatureTransformer(nn.Module):
    """A Local Feature Transformer (LoFTR) module."""

    def __init__(self, layer_names, topks=[16, 8, 8], d_model=256):
        super(LocalFeatureTransformer, self).__init__()
        self.d_model = d_model
        self.nhead = 8
        self.layer_names = layer_names

        encoder_layer = QuadtreeBlock(256, 8, attn_type="B", topks=topks, scale=3)
        self.layers = nn.ModuleList(
            [copy.deepcopy(encoder_layer) for _ in range(len(self.layer_names))]
        )

        self._reset_parameters()

    def _reset_parameters(self):
        for name, p in self.named_parameters():
            if "temp" in name or "radius_offset" in name:
                continue
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def forward(self, feat0, feat1, mask0=None, mask1=None, H=0, W=0, pos0=0, pos1=0):
        """
        Args:
            feat0 (torch.Tensor): [N, L, C]
            feat1 (torch.Tensor): [N, S, C]
            mask0 (torch.Tensor): [N, L] (optional)
            mask1 (torch.Tensor): [N, S] (optional)
        """
        assert self.d_model == feat0.size(
            2
        ), "the feature number of src and transformer must be equal"
        if len(feat0.shape) == 4:
            B, C, H, W = feat0.shape
            feat0 = rearrange(feat0, "b c h w -> b (h w) c")
            feat1 = rearrange(feat1, "b c h w -> b (h w) c")
        ii = -1
        for layer, name in zip(self.layers, self.layer_names):
            if name == "self":
                feat0 = layer(feat0, feat0, H, W)
                feat1 = layer(feat1, feat1, H, W)
            elif name == "cross":
                feat0, feat1 = layer(feat0, feat1, H, W), layer(feat1, feat0, H, W)
            else:
                raise KeyError
        return feat0, feat1


class PositionEncodingSineNorm(nn.Module):
    """
    This is a sinusoidal position encoding that generalized to 2-dimensional images
    """

    def __init__(self, d_model, max_shape=(256, 256)):
        """
        Args:
            max_shape (tuple): for 1/8 featmap, the max length of 256 corresponds to 2048 pixels
            temp_bug_fix (bool): As noted in this [issue](https://github.com/zju3dv/LoFTR/issues/41),
                the original implementation of LoFTR includes a bug in the pos-enc impl, which has little impact
                on the final performance. For now, we keep both impls for backward compatability.
                We will remove the buggy impl after re-training all variants of our released models.
        """
        super().__init__()
        self.d_model = d_model
        self.max_shape = max_shape

        pe = torch.zeros((d_model, *max_shape))
        y_position = torch.ones(max_shape).cumsum(0).float().unsqueeze(0)
        x_position = torch.ones(max_shape).cumsum(1).float().unsqueeze(0)

        div_term = torch.exp(
            torch.arange(0, d_model // 2, 2).float()
            * (-math.log(10000.0) / (d_model // 2))
        )
        div_term = div_term[:, None, None]  # [C//4, 1, 1]
        pe[0::4, :, :] = torch.sin(x_position * div_term)
        pe[1::4, :, :] = torch.cos(x_position * div_term)
        pe[2::4, :, :] = torch.sin(y_position * div_term)
        pe[3::4, :, :] = torch.cos(y_position * div_term)
        self.eval_reso = None

        self.register_buffer("pe", pe.unsqueeze(0), persistent=False)  # [1, C, H, W]

    def forward(self, x, train_reso=None, eval_reso=None):
        """
        Args:
            x: [N, C, H, W]
        """
        if self.pe.dtype != x.dtype:
            self.pe.to(dtype=x.dtype)

        if train_reso is None and eval_reso is None:
            return (
                x + self.pe[:, :, : x.size(2), : x.size(3)],
                self.pe[:, :, : x.size(2), : x.size(3)],
            )
        elif eval_reso != self.eval_reso:
            self.eval_reso = eval_reso
            pe = torch.zeros(
                (self.d_model, *self.max_shape), dtype=x.dtype, device=x.device
            )
            y_position = (
                torch.ones(self.max_shape, dtype=x.dtype, device=x.device)
                .cumsum(0)
                .unsqueeze(0)
                * train_reso[0]
                / eval_reso[0]
            )
            x_position = (
                torch.ones(self.max_shape, dtype=x.dtype, device=x.device)
                .cumsum(1)
                .unsqueeze(0)
                * train_reso[1]
                / eval_reso[1]
            )

            div_term = torch.exp(
                torch.arange(0, self.d_model // 2, 2, dtype=x.dtype, device=x.device)
                * (-math.log(10000.0) / (self.d_model // 2))
            )
            div_term = div_term[:, None, None]  # [C//4, 1, 1]
            pe[0::4, :, :] = torch.sin(x_position * div_term)
            pe[1::4, :, :] = torch.cos(x_position * div_term)
            pe[2::4, :, :] = torch.sin(y_position * div_term)
            pe[3::4, :, :] = torch.cos(y_position * div_term)
            pe = pe.unsqueeze(0).to(x.device)
            self.register_buffer("eval_pe", pe, persistent=False)  # [1, C, H, W]

            return (
                x + pe[:, :, : x.size(2), : x.size(3)],
                pe[:, :, : x.size(2), : x.size(3)],
            )
        else:
            if self.eval_pe.dtype != x.dtype:
                self.eval_pe.to(dtype=x.dtype)

            return (
                x + self.eval_pe[:, :, : x.size(2), : x.size(3)],
                self.eval_pe[:, :, : x.size(2), : x.size(3)],
            )


class MatchingModel(nn.Module):
    def __init__(self, train_size=(384, 512)):
        super().__init__()
        self.image_size = train_size
        self.backbone = ResNetFPN_8_2()

        self.pos_encoding = PositionEncodingSineNorm(256)

        self.loftr_coarse = LocalFeatureTransformer(
            layer_names=[
                "self",
                "cross",
                "self",
                "cross",
                "self",
                "cross",
                "self",
                "cross",
            ],
            topks=[16, 8, 8],
        )

    def forward(self, x):
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)
        # input x in [-1, 1]
        x = (x + 1) / 2.0
        feats_c = self.backbone(x)
        B, _, H, W = feats_c.shape

        (feat_c0, feat_c1) = feats_c.split(B // 2)

        # 2. coarse-level loftr module
        if self.training:
            feat_c0, pos_encoding0 = self.pos_encoding(feat_c0, None, None)
            feat_c1, pos_encoding1 = self.pos_encoding(feat_c1, None, None)
        else:
            feat_c0, pos_encoding0 = self.pos_encoding(
                feat_c0, self.image_size, x.shape[2:4]
            )
            feat_c1, pos_encoding1 = self.pos_encoding(
                feat_c1, self.image_size, x.shape[2:4]
            )

        feat_c0 = rearrange(feat_c0, "n c h w -> n (h w) c")
        feat_c1 = rearrange(feat_c1, "n c h w -> n (h w) c")

        mask_c0 = mask_c1 = None  # mask is useful in training
        feat_c0, feat_c1 = self.loftr_coarse(
            feat_c0,
            feat_c1,
            mask_c0,
            mask_c1,
            H=H,
            W=W,
            pos0=pos_encoding0,
            pos1=pos_encoding1,
        )

        feat_c0 = rearrange(feat_c0, "n (h w) c -> n c h w", h=H)
        feat_c1 = rearrange(feat_c1, "n (h w) c -> n c h w", h=H)

        output = torch.cat([feat_c0, feat_c1], dim=0)

        if is_list:
            output = torch.split(output, [batch_dim, batch_dim], dim=0)

        return output

    def load_state_dict(self, path=""):
        state_dict = torch.load(path, map_location="cpu")["state_dict"]
        for k in list(state_dict.keys()):
            if k.startswith("matcher."):
                if "radius_offset" in k:
                    state_dict.pop(k)
                else:
                    state_dict[k.replace("matcher.", "", 1)] = state_dict.pop(k)

        for k in list(state_dict.keys()):
            if (
                k.startswith("backbone.")
                or k.startswith("pos_encoding.")
                or k.startswith("loftr_coarse.")
            ):
                continue
            else:
                state_dict.pop(k)

        torch_init_model(self, state_dict, key="model")
        self.backbone.reset_model()
        return self

    def reset_model(self):
        self.backbone.reset_model()
        return self



import json
import logging
import os
from functools import partial
from typing import Union, Optional

import torch
from torch.hub import (
    load_state_dict_from_url,
    download_url_to_file,
    urlparse,
    HASH_REGEX,
)

try:
    from torch.hub import get_dir
except ImportError:
    from torch.hub import _get_torch_home as get_dir

from timm import __version__

try:
    from huggingface_hub import hf_hub_url
    from huggingface_hub import cached_download

    cached_download = partial(
        cached_download, library_name="timm", library_version=__version__
    )
except ImportError:
    hf_hub_url = None
    cached_download = None

_logger = logging.getLogger(__name__)


def get_cache_dir(child_dir=""):
    """
    Returns the location of the directory where models are cached (and creates it if necessary).
    """
    # Issue warning to move data if old env is set
    if os.getenv("TORCH_MODEL_ZOO"):
        _logger.warning(
            "TORCH_MODEL_ZOO is deprecated, please use env TORCH_HOME instead"
        )

    hub_dir = get_dir()
    child_dir = () if not child_dir else (child_dir,)
    model_dir = os.path.join(hub_dir, "checkpoints", *child_dir)
    os.makedirs(model_dir, exist_ok=True)
    return model_dir


def download_cached_file(url, check_hash=True, progress=False):
    parts = urlparse(url)
    filename = os.path.basename(parts.path)
    cached_file = os.path.join(get_cache_dir(), filename)
    if not os.path.exists(cached_file):
        _logger.info('Downloading: "{}" to {}\n'.format(url, cached_file))
        hash_prefix = None
        if check_hash:
            r = HASH_REGEX.search(filename)  # r is Optional[Match[str]]
            hash_prefix = r.group(1) if r else None
        download_url_to_file(url, cached_file, hash_prefix, progress=progress)
    return cached_file


def has_hf_hub(necessary=False):
    if hf_hub_url is None and necessary:
        # if no HF Hub module installed and it is necessary to continue, raise error
        raise RuntimeError(
            "Hugging Face hub model specified but package not installed. Run `pip install huggingface_hub`."
        )
    return hf_hub_url is not None


def hf_split(hf_id):
    rev_split = hf_id.split("@")
    assert (
        0 < len(rev_split) <= 2
    ), "hf_hub id should only contain one @ character to identify revision."
    hf_model_id = rev_split[0]
    hf_revision = rev_split[-1] if len(rev_split) > 1 else None
    return hf_model_id, hf_revision


def load_cfg_from_json(json_file: Union[str, os.PathLike]):
    with open(json_file, "r", encoding="utf-8") as reader:
        text = reader.read()
    return json.loads(text)


def _download_from_hf(model_id: str, filename: str):
    hf_model_id, hf_revision = hf_split(model_id)
    url = hf_hub_url(hf_model_id, filename, revision=hf_revision)
    return cached_download(url, cache_dir=get_cache_dir("hf"))


def load_model_config_from_hf(model_id: str):
    assert has_hf_hub(True)
    cached_file = _download_from_hf(model_id, "config.json")
    default_cfg = load_cfg_from_json(cached_file)
    default_cfg["hf_hub"] = (
        model_id  # insert hf_hub id for pretrained weight load during model creation
    )
    model_name = default_cfg.get("architecture")
    return default_cfg, model_name


def load_state_dict_from_hf(model_id: str):
    assert has_hf_hub(True)
    cached_file = _download_from_hf(model_id, "pytorch_model.bin")
    state_dict = torch.load(cached_file, map_location="cpu")
    return state_dict



""" Conv2d w/ Same Padding

Hacked together by / Copyright 2020 Ross Wightman
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Tuple, Optional

from .padding import pad_same, get_padding_value


def conv2d_same(
    x,
    weight: torch.Tensor,
    bias: Optional[torch.Tensor] = None,
    stride: Tuple[int, int] = (1, 1),
    padding: Tuple[int, int] = (0, 0),
    dilation: Tuple[int, int] = (1, 1),
    groups: int = 1,
):
    x = pad_same(x, weight.shape[-2:], stride, dilation)
    return F.conv2d(x, weight, bias, stride, (0, 0), dilation, groups)


class Conv2dSame(nn.Conv2d):
    """Tensorflow like 'SAME' convolution wrapper for 2D convolutions"""

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
    ):
        super(Conv2dSame, self).__init__(
            in_channels, out_channels, kernel_size, stride, 0, dilation, groups, bias
        )

    def forward(self, x):
        return conv2d_same(
            x,
            self.weight,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )


def create_conv2d_pad(in_chs, out_chs, kernel_size, **kwargs):
    padding = kwargs.pop("padding", "")
    kwargs.setdefault("bias", False)
    padding, is_dynamic = get_padding_value(padding, kernel_size, **kwargs)
    if is_dynamic:
        return Conv2dSame(in_chs, out_chs, kernel_size, **kwargs)
    else:
        return nn.Conv2d(in_chs, out_chs, kernel_size, padding=padding, **kwargs)



from __future__ import print_function
import torch
import torch.nn as nn
import torch.utils.data
from torch.autograd import Variable
import torch.nn.functional as F
import math
import numpy as np
import pdb


class residualBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        in_channels,
        n_filters,
        stride=1,
        downsample=None,
        dilation=1,
        with_bn=True,
    ):
        super(residualBlock, self).__init__()
        if dilation > 1:
            padding = dilation
        else:
            padding = 1

        if with_bn:
            self.convbnrelu1 = conv2DBatchNormRelu(
                in_channels, n_filters, 3, stride, padding, dilation=dilation
            )
            self.convbn2 = conv2DBatchNorm(n_filters, n_filters, 3, 1, 1)
        else:
            self.convbnrelu1 = conv2DBatchNormRelu(
                in_channels,
                n_filters,
                3,
                stride,
                padding,
                dilation=dilation,
                with_bn=False,
            )
            self.convbn2 = conv2DBatchNorm(n_filters, n_filters, 3, 1, 1, with_bn=False)
        self.downsample = downsample
        self.relu = nn.LeakyReLU(0.1, inplace=True)

    def forward(self, x):
        residual = x

        out = self.convbnrelu1(x)
        out = self.convbn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        return self.relu(out)


def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=True,
        ),
        nn.BatchNorm2d(out_planes),
        nn.LeakyReLU(0.1, inplace=True),
    )


class conv2DBatchNorm(nn.Module):
    def __init__(
        self, in_channels, n_filters, k_size, stride, padding, dilation=1, with_bn=True
    ):
        super(conv2DBatchNorm, self).__init__()
        bias = not with_bn

        if dilation > 1:
            conv_mod = nn.Conv2d(
                int(in_channels),
                int(n_filters),
                kernel_size=k_size,
                padding=padding,
                stride=stride,
                bias=bias,
                dilation=dilation,
            )

        else:
            conv_mod = nn.Conv2d(
                int(in_channels),
                int(n_filters),
                kernel_size=k_size,
                padding=padding,
                stride=stride,
                bias=bias,
                dilation=1,
            )

        if with_bn:
            self.cb_unit = nn.Sequential(
                conv_mod,
                nn.BatchNorm2d(int(n_filters)),
            )
        else:
            self.cb_unit = nn.Sequential(
                conv_mod,
            )

    def forward(self, inputs):
        outputs = self.cb_unit(inputs)
        return outputs


class conv2DBatchNormRelu(nn.Module):
    def __init__(
        self, in_channels, n_filters, k_size, stride, padding, dilation=1, with_bn=True
    ):
        super(conv2DBatchNormRelu, self).__init__()
        bias = not with_bn
        if dilation > 1:
            conv_mod = nn.Conv2d(
                int(in_channels),
                int(n_filters),
                kernel_size=k_size,
                padding=padding,
                stride=stride,
                bias=bias,
                dilation=dilation,
            )

        else:
            conv_mod = nn.Conv2d(
                int(in_channels),
                int(n_filters),
                kernel_size=k_size,
                padding=padding,
                stride=stride,
                bias=bias,
                dilation=1,
            )

        if with_bn:
            self.cbr_unit = nn.Sequential(
                conv_mod,
                nn.BatchNorm2d(int(n_filters)),
                nn.LeakyReLU(0.1, inplace=True),
            )
        else:
            self.cbr_unit = nn.Sequential(
                conv_mod,
                nn.LeakyReLU(0.1, inplace=True),
            )

    def forward(self, inputs):
        outputs = self.cbr_unit(inputs)
        return outputs


class pyramidPooling(nn.Module):
    def __init__(self, in_channels, with_bn=True, levels=4):
        super(pyramidPooling, self).__init__()
        self.levels = levels

        self.paths = []
        for i in range(levels):
            self.paths.append(
                conv2DBatchNormRelu(in_channels, in_channels, 1, 1, 0, with_bn=with_bn)
            )
        self.path_module_list = nn.ModuleList(self.paths)
        self.relu = nn.LeakyReLU(0.1, inplace=True)

    def forward(self, x):
        h, w = x.shape[2:]

        k_sizes = []
        strides = []
        for pool_size in np.linspace(1, min(h, w) // 2, self.levels, dtype=int):
            k_sizes.append((int(h / pool_size), int(w / pool_size)))
            strides.append((int(h / pool_size), int(w / pool_size)))
        k_sizes = k_sizes[::-1]
        strides = strides[::-1]

        pp_sum = x

        for i, module in enumerate(self.path_module_list):
            out = F.avg_pool2d(x, k_sizes[i], stride=strides[i], padding=0)
            out = module(out)
            out = F.interpolate(out, size=(h, w), mode="bilinear", align_corners=False)
            pp_sum = pp_sum + 1.0 / self.levels * out
        pp_sum = self.relu(pp_sum / 2.0)

        return pp_sum


class pspnet(nn.Module):
    """
    Modified PSPNet.  https://github.com/meetshah1995/pytorch-semseg/blob/master/ptsemseg/models/pspnet.py
    """

    def __init__(self, is_proj=True, groups=1):
        super(pspnet, self).__init__()
        self.inplanes = 32
        self.is_proj = is_proj

        # Encoder
        self.convbnrelu1_1 = conv2DBatchNormRelu(
            in_channels=3, k_size=3, n_filters=16, padding=1, stride=2
        )
        self.convbnrelu1_2 = conv2DBatchNormRelu(
            in_channels=16, k_size=3, n_filters=16, padding=1, stride=1
        )
        self.convbnrelu1_3 = conv2DBatchNormRelu(
            in_channels=16, k_size=3, n_filters=32, padding=1, stride=1
        )
        # Vanilla Residual Blocks
        self.res_block3 = self._make_layer(residualBlock, 64, 1, stride=2)
        self.res_block5 = self._make_layer(residualBlock, 128, 1, stride=2)
        self.res_block6 = self._make_layer(residualBlock, 128, 1, stride=2)
        self.res_block7 = self._make_layer(residualBlock, 128, 1, stride=2)
        self.pyramid_pooling = pyramidPooling(128, levels=3)

        # Iconvs
        self.upconv6 = nn.Sequential(
            nn.Upsample(scale_factor=2, align_corners=False),
            conv2DBatchNormRelu(
                in_channels=128, k_size=3, n_filters=64, padding=1, stride=1
            ),
        )
        self.iconv5 = conv2DBatchNormRelu(
            in_channels=192, k_size=3, n_filters=128, padding=1, stride=1
        )
        self.upconv5 = nn.Sequential(
            nn.Upsample(scale_factor=2, align_corners=False),
            conv2DBatchNormRelu(
                in_channels=128, k_size=3, n_filters=64, padding=1, stride=1
            ),
        )
        self.iconv4 = conv2DBatchNormRelu(
            in_channels=192, k_size=3, n_filters=128, padding=1, stride=1
        )
        self.upconv4 = nn.Sequential(
            nn.Upsample(scale_factor=2, align_corners=False),
            conv2DBatchNormRelu(
                in_channels=128, k_size=3, n_filters=64, padding=1, stride=1
            ),
        )
        self.iconv3 = conv2DBatchNormRelu(
            in_channels=128, k_size=3, n_filters=64, padding=1, stride=1
        )
        self.upconv3 = nn.Sequential(
            nn.Upsample(scale_factor=2, align_corners=False),
            conv2DBatchNormRelu(
                in_channels=64, k_size=3, n_filters=32, padding=1, stride=1
            ),
        )
        self.iconv2 = conv2DBatchNormRelu(
            in_channels=64, k_size=3, n_filters=64, padding=1, stride=1
        )

        if self.is_proj:
            self.proj6 = conv2DBatchNormRelu(
                in_channels=128, k_size=1, n_filters=128 // groups, padding=0, stride=1
            )
            self.proj5 = conv2DBatchNormRelu(
                in_channels=128, k_size=1, n_filters=128 // groups, padding=0, stride=1
            )
            self.proj4 = conv2DBatchNormRelu(
                in_channels=128, k_size=1, n_filters=128 // groups, padding=0, stride=1
            )
            self.proj3 = conv2DBatchNormRelu(
                in_channels=64, k_size=1, n_filters=64 // groups, padding=0, stride=1
            )
            self.proj2 = conv2DBatchNormRelu(
                in_channels=64, k_size=1, n_filters=64 // groups, padding=0, stride=1
            )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
                if hasattr(m.bias, "data"):
                    m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                nn.BatchNorm2d(planes * block.expansion),
            )
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
        return nn.Sequential(*layers)

    def forward(self, x):
        # H, W -> H/2, W/2
        conv1 = self.convbnrelu1_1(x)
        conv1 = self.convbnrelu1_2(conv1)
        conv1 = self.convbnrelu1_3(conv1)

        ## H/2, W/2 -> H/4, W/4
        pool1 = F.max_pool2d(conv1, 3, 2, 1)

        # H/4, W/4 -> H/16, W/16
        rconv3 = self.res_block3(pool1)
        conv4 = self.res_block5(rconv3)
        conv5 = self.res_block6(conv4)
        conv6 = self.res_block7(conv5)
        conv6 = self.pyramid_pooling(conv6)

        conv6x = F.interpolate(
            conv6,
            [conv5.size()[2], conv5.size()[3]],
            mode="bilinear",
            align_corners=False,
        )
        concat5 = torch.cat((conv5, self.upconv6[1](conv6x)), dim=1)
        conv5 = self.iconv5(concat5)

        conv5x = F.interpolate(
            conv5,
            [conv4.size()[2], conv4.size()[3]],
            mode="bilinear",
            align_corners=False,
        )
        concat4 = torch.cat((conv4, self.upconv5[1](conv5x)), dim=1)
        conv4 = self.iconv4(concat4)

        conv4x = F.interpolate(
            conv4,
            [rconv3.size()[2], rconv3.size()[3]],
            mode="bilinear",
            align_corners=False,
        )
        concat3 = torch.cat((rconv3, self.upconv4[1](conv4x)), dim=1)
        conv3 = self.iconv3(concat3)

        conv3x = F.interpolate(
            conv3,
            [pool1.size()[2], pool1.size()[3]],
            mode="bilinear",
            align_corners=False,
        )
        concat2 = torch.cat((pool1, self.upconv3[1](conv3x)), dim=1)
        conv2 = self.iconv2(concat2)

        if self.is_proj:
            proj6 = self.proj6(conv6)
            proj5 = self.proj5(conv5)
            proj4 = self.proj4(conv4)
            proj3 = self.proj3(conv3)
            proj2 = self.proj2(conv2)
            return proj6, proj5, proj4, proj3, proj2
        else:
            return conv6, conv5, conv4, conv3, conv2


class pspnet_s(nn.Module):
    """
    Modified PSPNet.  https://github.com/meetshah1995/pytorch-semseg/blob/master/ptsemseg/models/pspnet.py
    """

    def __init__(self, is_proj=True, groups=1):
        super(pspnet_s, self).__init__()
        self.inplanes = 32
        self.is_proj = is_proj

        # Encoder
        self.convbnrelu1_1 = conv2DBatchNormRelu(
            in_channels=3, k_size=3, n_filters=16, padding=1, stride=2
        )
        self.convbnrelu1_2 = conv2DBatchNormRelu(
            in_channels=16, k_size=3, n_filters=16, padding=1, stride=1
        )
        self.convbnrelu1_3 = conv2DBatchNormRelu(
            in_channels=16, k_size=3, n_filters=32, padding=1, stride=1
        )
        # Vanilla Residual Blocks
        self.res_block3 = self._make_layer(residualBlock, 64, 1, stride=2)
        self.res_block5 = self._make_layer(residualBlock, 128, 1, stride=2)
        self.res_block6 = self._make_layer(residualBlock, 128, 1, stride=2)
        self.res_block7 = self._make_layer(residualBlock, 128, 1, stride=2)
        self.pyramid_pooling = pyramidPooling(128, levels=3)

        # Iconvs
        self.upconv6 = nn.Sequential(
            nn.Upsample(scale_factor=2, align_corners=False),
            conv2DBatchNormRelu(
                in_channels=128, k_size=3, n_filters=64, padding=1, stride=1
            ),
        )
        self.iconv5 = conv2DBatchNormRelu(
            in_channels=192, k_size=3, n_filters=128, padding=1, stride=1
        )
        self.upconv5 = nn.Sequential(
            nn.Upsample(scale_factor=2, align_corners=False),
            conv2DBatchNormRelu(
                in_channels=128, k_size=3, n_filters=64, padding=1, stride=1
            ),
        )
        self.iconv4 = conv2DBatchNormRelu(
            in_channels=192, k_size=3, n_filters=128, padding=1, stride=1
        )
        self.upconv4 = nn.Sequential(
            nn.Upsample(scale_factor=2, align_corners=False),
            conv2DBatchNormRelu(
                in_channels=128, k_size=3, n_filters=64, padding=1, stride=1
            ),
        )
        self.iconv3 = conv2DBatchNormRelu(
            in_channels=128, k_size=3, n_filters=64, padding=1, stride=1
        )
        # self.upconv3 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
        #                             conv2DBatchNormRelu(in_channels=64, k_size=3, n_filters=32,
        #                                         padding=1, stride=1))
        # self.iconv2 = conv2DBatchNormRelu(in_channels=64, k_size=3, n_filters=64,
        #                                         padding=1, stride=1)

        if self.is_proj:
            self.proj6 = conv2DBatchNormRelu(
                in_channels=128, k_size=1, n_filters=128 // groups, padding=0, stride=1
            )
            self.proj5 = conv2DBatchNormRelu(
                in_channels=128, k_size=1, n_filters=128 // groups, padding=0, stride=1
            )
            self.proj4 = conv2DBatchNormRelu(
                in_channels=128, k_size=1, n_filters=128 // groups, padding=0, stride=1
            )
            self.proj3 = conv2DBatchNormRelu(
                in_channels=64, k_size=1, n_filters=64 // groups, padding=0, stride=1
            )
            # self.proj2 = conv2DBatchNormRelu(in_channels=64, k_size=1,n_filters=64//groups, padding=0,stride=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
                if hasattr(m.bias, "data"):
                    m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                nn.BatchNorm2d(planes * block.expansion),
            )
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
        return nn.Sequential(*layers)

    def forward(self, x):
        # H, W -> H/2, W/2
        conv1 = self.convbnrelu1_1(x)
        conv1 = self.convbnrelu1_2(conv1)
        conv1 = self.convbnrelu1_3(conv1)

        ## H/2, W/2 -> H/4, W/4
        pool1 = F.max_pool2d(conv1, 3, 2, 1)

        # H/4, W/4 -> H/16, W/16
        rconv3 = self.res_block3(pool1)
        conv4 = self.res_block5(rconv3)
        conv5 = self.res_block6(conv4)
        conv6 = self.res_block7(conv5)
        conv6 = self.pyramid_pooling(conv6)

        conv6x = F.interpolate(
            conv6,
            [conv5.size()[2], conv5.size()[3]],
            mode="bilinear",
            align_corners=False,
        )
        concat5 = torch.cat((conv5, self.upconv6[1](conv6x)), dim=1)
        conv5 = self.iconv5(concat5)

        conv5x = F.interpolate(
            conv5,
            [conv4.size()[2], conv4.size()[3]],
            mode="bilinear",
            align_corners=False,
        )
        concat4 = torch.cat((conv4, self.upconv5[1](conv5x)), dim=1)
        conv4 = self.iconv4(concat4)

        conv4x = F.interpolate(
            conv4,
            [rconv3.size()[2], rconv3.size()[3]],
            mode="bilinear",
            align_corners=False,
        )
        concat3 = torch.cat((rconv3, self.upconv4[1](conv4x)), dim=1)
        conv3 = self.iconv3(concat3)

        # conv3x = F.interpolate(conv3, [pool1.size()[2],pool1.size()[3]],mode='bilinear', align_corners=False)
        # concat2 = torch.cat((pool1,self.upconv3[1](conv3x)),dim=1)
        # conv2 = self.iconv2(concat2)

        if self.is_proj:
            proj6 = self.proj6(conv6)
            proj5 = self.proj5(conv5)
            proj4 = self.proj4(conv4)
            proj3 = self.proj3(conv3)
            #    proj2 = self.proj2(conv2)
            #    return proj6,proj5,proj4,proj3,proj2
            return proj6, proj5, proj4, proj3
        else:
            #    return conv6, conv5, conv4, conv3, conv2
            return conv6, conv5, conv4, conv3



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from .cnn import BasicEncoder
from .encoder import MemoryEncoder
from .encoders import twins_svt_large
from .decoder import MemoryDecoder
from .utils import compute_grid_indices, compute_weight
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""
        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid[:, None] * i_loss).mean()

        return flow_loss


class FlowFormer(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformer-chairs-84881320.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformer-things-dbe62dd3.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformer-sintel-cce498f8.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformer-kitti-d4225180.ckpt",
    }

    def __init__(
        self,
        add_flow_token: bool = True,
        cnet: str = "twins",
        context_concat: bool = False,
        cost_encoder_res: bool = True,
        cost_heads_num: int = 1,
        cost_latent_dim: int = 128,
        cost_latent_input_dim: int = 64,
        cost_latent_token_num: int = 8,
        decoder_depth: int = 32,
        dropout: float = 0.0,
        encoder_depth: int = 3,
        encoder_latent_dim: int = 256,
        feat_cross_attn: bool = False,
        fnet: str = "twins",
        gamma: float = 0.8,
        max_flow: float = 400.0,
        gma: bool = True,
        only_global: bool = False,
        patch_size: int = 8,
        pe: str = "linear",
        pretrain: bool = False,
        query_latent_dim: int = 64,
        use_mlp: bool = False,
        mlp_expansion_factor: int = 4,
        vert_c_dim: int = 64,
        vertical_conv: bool = False,
        cost_scale_aug: Optional[tuple[float, float]] = None,
        use_tile_input: bool = True,
        tile_height: int = 432,
        tile_sigma: float = 0.05,
        train_size: Optional[tuple[int, int]] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.add_flow_token = add_flow_token
        self.cnet = cnet
        self.context_concat = context_concat
        self.cost_encoder_res = cost_encoder_res
        self.cost_heads_num = cost_heads_num
        self.cost_latent_dim = cost_latent_dim
        self.cost_latent_input_dim = cost_latent_input_dim
        self.cost_latent_token_num = cost_latent_token_num
        self.decoder_depth = decoder_depth
        self.dropout = dropout
        self.encoder_depth = encoder_depth
        self.encoder_latent_dim = encoder_latent_dim
        self.feat_cross_attn = feat_cross_attn
        self.fnet = fnet
        self.gma = gma
        self.only_global = only_global
        self.patch_size = patch_size
        self.pe = pe
        self.pretrain = pretrain
        self.query_latent_dim = query_latent_dim
        self.use_mlp = use_mlp
        self.mlp_expansion_factor = mlp_expansion_factor
        self.vert_c_dim = vert_c_dim
        self.vertical_conv = vertical_conv
        self.cost_scale_aug = cost_scale_aug
        self.use_tile_input = use_tile_input
        self.tile_height = tile_height
        self.tile_sigma = tile_sigma
        self.train_size = train_size

        if self.gma is None:
            self.gma = True  # Use GMA by default, unless

        self.memory_encoder = MemoryEncoder(
            fnet,
            encoder_latent_dim=encoder_latent_dim,
            pretrain=pretrain,
            feat_cross_attn=feat_cross_attn,
            cost_heads_num=cost_heads_num,
            patch_size=patch_size,
            cost_latent_input_dim=cost_latent_input_dim,
            pe=pe,
            encoder_depth=encoder_depth,
            cost_latent_dim=cost_latent_dim,
            dropout=dropout,
            use_mlp=use_mlp,
            cost_scale_aug=cost_scale_aug,
            mlp_expansion_factor=mlp_expansion_factor,
            vert_c_dim=vert_c_dim,
            vertical_conv=vertical_conv,
            cost_latent_token_num=cost_latent_token_num,
            cost_encoder_res=cost_encoder_res,
        )
        self.memory_decoder = MemoryDecoder(
            query_latent_dim=query_latent_dim,
            cost_heads_num=cost_heads_num,
            decoder_depth=decoder_depth,
            gma=gma,
            only_global=only_global,
            patch_size=patch_size,
            cost_latent_dim=cost_latent_dim,
            add_flow_token=add_flow_token,
            dropout=dropout,
        )
        if cnet == "twins":
            self.context_encoder = twins_svt_large(pretrained=self.pretrain)
        elif cnet == "basicencoder":
            self.context_encoder = BasicEncoder(output_dim=256, norm_fn="instance")

        self.showed_warning = False

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        if self.train_size is not None:
            train_size = self.train_size
        else:
            train_size = self.train_size

        if self.use_tile_input and train_size is None and not self.showed_warning:
            print(
                "WARNING: --train_size is not provided and it cannot be loaded from the checkpoint either. Flowformer will run without input tile."
            )
            self.showed_warning = True

        if self.use_tile_input and train_size is not None:
            return self.forward_tile(inputs, train_size)
        else:
            return self.forward_pad(inputs)

    def forward_pad(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        prev_flow = None
        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            prev_flow = inputs["prev_preds"]["flow_small"]

        flow_predictions, flow_small = self.predict(
            images[:, 0], images[:, 1], prev_flow
        )
        output_flow = flow_predictions[-1]

        if self.training:
            for i, p in enumerate(flow_predictions):
                flow_predictions[i] = self.postprocess_predictions(
                    p, image_resizer, is_flow=True
                )
            outputs = {
                "flows": flow_predictions[-1][:, None],
                "flow_preds": flow_predictions,
            }
        else:
            output_flow = self.postprocess_predictions(
                output_flow, image_resizer, is_flow=True
            )
            outputs = {"flows": output_flow[:, None], "flow_small": flow_small}

        return outputs

    def forward_tile(self, inputs, train_size):
        input_size = inputs["images"].shape[-2:]
        image_size = (max(self.tile_height, input_size[-2]), input_size[-1])
        hws = compute_grid_indices(image_size, train_size)
        device = inputs["images"].device
        weights = compute_weight(
            hws, image_size, train_size, self.tile_sigma, device=device
        )

        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            target_size=image_size,
            pad_two_side=False,
            pad_mode="constant",
            pad_value=-1,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        flows = 0
        flow_count = 0

        for idx, (h, w) in enumerate(hws):
            image1_tile = image1[:, :, h : h + train_size[0], w : w + train_size[1]]
            image2_tile = image2[:, :, h : h + train_size[0], w : w + train_size[1]]

            flow_predictions, _ = self.predict(image1_tile, image2_tile)
            flow_pre = flow_predictions[-1]

            padding = (
                w,
                image_size[1] - w - train_size[1],
                h,
                image_size[0] - h - train_size[0],
                0,
                0,
            )
            flows += F.pad(flow_pre * weights[idx], padding)
            flow_count += F.pad(weights[idx], padding)

        output_flow = flows / flow_count

        output_flow = self.postprocess_predictions(
            output_flow, image_resizer, is_flow=True
        )
        return {"flows": output_flow[:, None]}

    def predict(self, image1, image2, prev_flow=None):
        """Estimate optical flow between pair of frames"""

        data = {}

        if self.context_concat:
            context = self.context_encoder(torch.cat([image1, image2], dim=1))
        else:
            context = self.context_encoder(image1)

        cost_memory = self.memory_encoder(image1, image2, data, context)

        flow_predictions, flow_small = self.memory_decoder(
            cost_memory, context, data, prev_flow=prev_flow
        )

        return flow_predictions, flow_small


@register_model
@trainable
class flowformer(FlowFormer):
    pass



# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from pathlib import Path
import shutil

from jsonargparse import ArgumentParser

import ptlflow
from ptlflow.data.flow_datamodule import FlowDataModule
from ptlflow.utils.dummy_datasets import write_kitti
import test

TEST_MODEL = "raft_small"


def test_test(tmp_path: Path) -> None:
    model = ptlflow.get_model(TEST_MODEL)

    data_parser = ArgumentParser()
    data_parser.add_class_arguments(FlowDataModule, "data")
    data_args = data_parser.parse_args([])
    data_args.data.test_dataset = "kitti-2015"
    data_args.data.kitti_2015_root_dir = str(tmp_path / "KITTI/2015")

    data_parser = ArgumentParser(exit_on_error=False)
    data_parser.add_argument("--data", type=FlowDataModule)
    data_cfg = data_parser.parse_object({"data": data_args.data})
    datamodule = data_parser.instantiate_classes(data_cfg).data
    datamodule.setup("test")

    parser = ArgumentParser(parents=[test._init_parser()])
    args = parser.parse_args([])
    args.output_path = str(tmp_path)

    write_kitti(tmp_path)

    test.test(args, model, datamodule)

    dataset_name_path = [
        ("kitti2015", "flow/000000_10.png"),
    ]
    for dname, dpath in dataset_name_path:
        print(tmp_path / dname / dpath)
        assert (tmp_path / dname / dpath).exists()

    shutil.rmtree(tmp_path)



""" Image to Patch Embedding using Conv2d

A convolution based approach to patchifying a 2D image w/ embedding projection.

Based on the impl in https://github.com/google-research/vision_transformer

Hacked together by / Copyright 2020 Ross Wightman
"""

from torch import nn as nn

from .helpers import to_2tuple


class PatchEmbed(nn.Module):
    """2D Image to Patch Embedding"""

    def __init__(
        self,
        img_size=224,
        patch_size=16,
        in_chans=3,
        embed_dim=768,
        norm_layer=None,
        flatten=True,
    ):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.grid_size = (img_size[0] // patch_size[0], img_size[1] // patch_size[1])
        self.num_patches = self.grid_size[0] * self.grid_size[1]
        self.flatten = flatten

        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        B, C, H, W = x.shape
        assert (
            H == self.img_size[0] and W == self.img_size[1]
        ), f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        x = self.proj(x)
        if self.flatten:
            x = x.flatten(2).transpose(1, 2)  # BCHW -> BNC
        x = self.norm(x)
        return x



# =============================================================================
# Copyright 2025 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from .norm import GroupNorm2d, BatchNorm2d, LayerNorm2d


def conv2d(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: Optional[torch.Tensor] = None,
    stride: int = 1,
    padding: int = 0,
    dilation: int = 1,
    groups: int = 1,
    pre_act_fn=None,
    post_act_fn=None,
    pre_norm=None,
    post_norm=None,
    is_transpose: bool = False,
) -> torch.Tensor:
    if pre_norm is not None:
        x = pre_norm(x)

    if pre_act_fn is not None:
        x = pre_act_fn(x)

    if is_transpose:
        x = F.conv_transpose2d(
            x,
            weight=weight,
            bias=bias,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
        )
    else:
        x = F.conv2d(
            x,
            weight=weight,
            bias=bias,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
        )

    if post_norm is not None:
        x = post_norm(x)

    if post_act_fn is not None:
        x = post_act_fn(x)

    return x


class ConvBase(nn.Module):
    def __init__(
        self,
        is_transpose,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        device=None,
        dtype=None,
        pre_norm=None,
        post_norm=None,
        group_norm_groups=8,
        pre_act_fn=None,
        post_act_fn=None,
    ) -> None:
        super().__init__()
        self.is_transpose = is_transpose
        self.in_channels = in_channels
        self.out_channels = out_channels
        if not isinstance(kernel_size, (list, tuple)):
            kernel_size = (kernel_size, kernel_size)
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.padding_mode = padding_mode
        self.device = device
        self.dtype = dtype
        self.pre_norm = pre_norm
        self.post_norm = post_norm
        self.pre_act_fn = pre_act_fn
        self.post_act_fn = post_act_fn

        if self.is_transpose:
            self.register_parameter(
                "weight",
                nn.Parameter(
                    torch.zeros(
                        in_channels // groups,
                        out_channels,
                        kernel_size[0],
                        kernel_size[1],
                        device=device,
                        dtype=dtype,
                    )
                ),
            )
        else:
            self.register_parameter(
                "weight",
                nn.Parameter(
                    torch.zeros(
                        out_channels,
                        in_channels // groups,
                        kernel_size[0],
                        kernel_size[1],
                        device=device,
                        dtype=dtype,
                    )
                ),
            )

        if bias:
            self.register_parameter(
                "bias",
                nn.Parameter(torch.zeros(out_channels, device=device, dtype=dtype)),
            )
        else:
            self.bias = None

        if isinstance(pre_norm, str):
            if pre_norm == "instance":
                self.pre_norm = nn.InstanceNorm2d(1, track_running_stats=False)
            elif pre_norm == "group":
                self.pre_norm = GroupNorm2d(
                    group_norm_groups, out_channels, affine=False
                )
            elif pre_norm == "batch":
                self.pre_norm = BatchNorm2d(out_channels)
            elif pre_norm == "layer":
                self.pre_norm = LayerNorm2d(out_channels)
        else:
            self.pre_norm = pre_norm

        if isinstance(post_norm, str):
            if post_norm == "instance":
                self.post_norm = nn.InstanceNorm2d(1, track_running_stats=False)
            elif post_norm == "group":
                self.post_norm = GroupNorm2d(
                    group_norm_groups, out_channels, affine=False
                )
            elif post_norm == "batch":
                self.post_norm = BatchNorm2d(out_channels)
            elif post_norm == "layer":
                self.post_norm = LayerNorm2d(out_channels)
        else:
            self.post_norm = post_norm

        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(k), 1/sqrt(k)), where k = weight.size(1) * prod(*kernel_size)
        # For more details see: https://github.com/pytorch/pytorch/issues/15314#issuecomment-477448573
        nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight)
            if fan_in != 0:
                bound = 1 / math.sqrt(fan_in)
                nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return conv2d(
            x=x,
            weight=self.weight,
            bias=self.bias,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            groups=self.groups,
            pre_act_fn=self.pre_act_fn,
            post_act_fn=self.post_act_fn,
            pre_norm=self.pre_norm,
            post_norm=self.post_norm,
            is_transpose=self.is_transpose,
        )


class Conv2dBlock(ConvBase):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        device=None,
        dtype=None,
        pre_norm=None,
        post_norm=None,
        group_norm_groups=8,
        pre_act_fn=None,
        post_act_fn=None,
    ) -> None:
        super().__init__(
            is_transpose=False,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            device=device,
            dtype=dtype,
            pre_norm=pre_norm,
            post_norm=post_norm,
            group_norm_groups=group_norm_groups,
            pre_act_fn=pre_act_fn,
            post_act_fn=post_act_fn,
        )


class ConvTranspose2dBlock(ConvBase):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        device=None,
        dtype=None,
        pre_norm=None,
        post_norm=None,
        group_norm_groups=8,
        pre_act_fn=None,
        post_act_fn=None,
    ) -> None:
        super().__init__(
            is_transpose=True,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            device=device,
            dtype=dtype,
            pre_norm=pre_norm,
            post_norm=post_norm,
            group_norm_groups=group_norm_groups,
            pre_act_fn=pre_act_fn,
            post_act_fn=post_act_fn,
        )



from torch.nn.modules.module import Module
import torch
import numpy as np
from torch.autograd import Variable
from ..functions import *
import torch.nn.functional as F

from ..functions.GANet import MyLossFunction
from ..functions.GANet import SgaFunction
from ..functions.GANet import NlfFunction
from ..functions.GANet import LgaFunction
from ..functions.GANet import Lga2Function
from ..functions.GANet import Lga3Function
from ..functions.GANet import Lga3dFunction
from ..functions.GANet import Lga3d2Function
from ..functions.GANet import Lga3d3Function
from ..functions.GANet import MyLoss2Function
from ..functions.GANet import NlfDownFunction
from ..functions.GANet import NlfUpFunction
from ..functions.GANet import NlfRightFunction
from ..functions.GANet import NlfLeftFunction


class ChannelNorm(Module):
    def __init__(self, eps=1e-5):
        super(ChannelNorm, self).__init__()
        #        self.weight = nn.Parameter(torch.ones(1,num_features,1,1))
        #        self.bias = nn.Parameter(torch.zeros(1,num_features,1,1))
        #        self.num_groups = num_groups
        self.eps = eps

    def forward(self, x):
        #        N,C,H,W = x.size()
        #        G = self.num_groups

        #        x = x.view(N,G,-1)
        mean = x.mean(1, keepdim=True)
        var = x.var(1, keepdim=True)

        x = (x - mean) / (var + self.eps).sqrt()
        return x


#        x = x.view(N,C,H,W)
#        return x * self.weight + self.bias


class GetWeights(Module):
    def __init__(self, wsize=5):
        super(GetWeights, self).__init__()
        self.wsize = wsize

    #        self.disp = Variable(torch.Tensor(np.reshape(np.array(range(self.maxdisp)),[1,self.maxdisp,1,1])).cuda(), requires_grad=False)
    def forward(self, x):
        assert x.is_contiguous() == True
        with torch.cuda.device_of(x):
            #            x = F.normalize(x, p=2, dim=1)
            num, channels, height, width = x.size()

            weight_down = x.new().resize_(num, 5, height, width).zero_()
            weight_down[:, 0, :, :] = torch.sum(x * x, 1)
            weight_down[:, 1, 1:, :] = torch.sum(x[:, :, 1:, :] * x[:, :, :-1, :], 1)
            weight_down[:, 2, 1:, 1:] = torch.sum(
                x[:, :, 1:, 1:] * x[:, :, :-1, :-1], 1
            )
            weight_down[:, 3, 1:, :-1] = torch.sum(
                x[:, :, 1:, :-1] * x[:, :, :-1, 1:], 1
            )
            weight_down[:, 4, :, 1:] = torch.sum(x[:, :, :, 1:] * x[:, :, :, :-1], 1)

            weight_up = x.new().resize_(num, 5, height, width).zero_()
            weight_up[:, 0, :, :] = torch.sum(x * x, 1)
            weight_up[:, 1, :-1, :] = torch.sum(x[:, :, :-1, :] * x[:, :, 1:, :], 1)
            weight_up[:, 2, :-1, 1:] = torch.sum(x[:, :, :-1, 1:] * x[:, :, 1:, :-1], 1)
            weight_up[:, 3, :-1, :-1] = torch.sum(
                x[:, :, :-1, :-1] * x[:, :, 1:, 1:], 1
            )
            weight_up[:, 4, :, :-1] = torch.sum(x[:, :, :, :-1] * x[:, :, :, 1:], 1)

            weight_right = x.new().resize_(num, 5, height, width).zero_()
            weight_right[:, 0, :, :] = torch.sum(x * x, 1)
            weight_right[:, 1, :, 1:] = torch.sum(x[:, :, :, 1:] * x[:, :, :, :-1], 1)
            weight_right[:, 2, 1:, 1:] = torch.sum(
                x[:, :, 1:, 1:] * x[:, :, :-1, :-1], 1
            )
            weight_right[:, 3, :-1, 1:] = torch.sum(
                x[:, :, :-1, 1:] * x[:, :, 1:, :-1], 1
            )
            weight_right[:, 4, 1:, :] = torch.sum(x[:, :, 1:, :] * x[:, :, :-1, :], 1)

            weight_left = x.new().resize_(num, 5, height, width).zero_()
            weight_left[:, 0, :, :] = torch.sum(x * x, 1)
            weight_left[:, 1, :, :-1] = torch.sum(x[:, :, :, :-1] * x[:, :, :, 1:], 1)
            weight_left[:, 2, 1:, :-1] = torch.sum(
                x[:, :, 1:, :-1] * x[:, :, :-1, 1:], 1
            )
            weight_left[:, 3, :-1, :-1] = torch.sum(
                x[:, :, :-1, :-1] * x[:, :, 1:, 1:], 1
            )
            weight_left[:, 4, :-1, :] = torch.sum(x[:, :, :-1, :] * x[:, :, 1:, :], 1)

            #            weight_down = F.normalize(weight_down, p=1, dim=1)
            #            weight_up = F.normalize(weight_up, p=1, dim=1)
            #            weight_right = F.normalize(weight_right, p=1, dim=1)
            #            weight_left = F.normalize(weight_left, p=1, dim=1)

            weight_down = F.softmax(weight_down, dim=1)
            weight_up = F.softmax(weight_up, dim=1)
            weight_right = F.softmax(weight_right, dim=1)
            weight_left = F.softmax(weight_left, dim=1)

            weight_down = weight_down.contiguous()
            weight_up = weight_up.contiguous()
            weight_right = weight_right.contiguous()
            weight_left = weight_left.contiguous()

        return weight_down, weight_up, weight_right, weight_left


class GetFilters(Module):
    def __init__(self, radius=2):
        super(GetFilters, self).__init__()
        self.radius = radius
        self.wsize = (radius * 2 + 1) * (radius * 2 + 1)

    #        self.disp = Variable(torch.Tensor(np.reshape(np.array(range(self.maxdisp)),[1,self.maxdisp,1,1])).cuda(), requires_grad=False)
    def forward(self, x):
        # assert(x.is_contiguous() == True)

        with torch.cuda.device_of(x):
            x = F.normalize(x, p=2, dim=1)
            #            num, channels, height, width = x.size()
            #            rem = torch.unsqueeze(x, 2).repeat(1, 1, self.wsize, 1, 1)
            #
            #            temp = x.new().resize_(num, channels, self.wsize, height, width).zero_()
            #            idx = 0
            #           for r in range(-self.radius, self.radius+1):
            #               for c in range(-self.radius, self.radius+1):
            #                   temp[:, :, idx, max(-r, 0):min(height - r, height), max(-c,0):min(width-c, width)] = x[:, :, max(r, 0):min(height + r, height), max(c, 0):min(width + c, width)]
            #                    idx += 1
            #            filters = torch.squeeze(torch.sum(rem*temp, 1), 1)
            #            filters = F.normalize(filters, p=1, dim=1)
            #            filters = filters.contiguous()
            #        return filters

            num, channels, height, width = x.size()
            filters = x.new().resize_(num, self.wsize, height, width).zero_()
            idx = 0
            for r in range(-self.radius, self.radius + 1):
                for c in range(-self.radius, self.radius + 1):
                    filters[
                        :,
                        idx,
                        max(-r, 0) : min(height - r, height),
                        max(-c, 0) : min(width - c, width),
                    ] = torch.squeeze(
                        torch.sum(
                            x[
                                :,
                                :,
                                max(r, 0) : min(height + r, height),
                                max(c, 0) : min(width + c, width),
                            ]
                            * x[
                                :,
                                :,
                                max(-r, 0) : min(height - r, height),
                                max(-c, 0) : min(width - c, width),
                            ],
                            1,
                        ),
                        1,
                    )
                    idx += 1

            filters = F.normalize(filters, p=1, dim=1)
            filters = filters.contiguous()
        return filters


class MyNormalize(Module):
    def __init__(self, dim):
        self.dim = dim
        super(MyNormalize, self).__init__()

    def forward(self, x):
        #        assert(x.is_contiguous() == True)
        with torch.cuda.device_of(x):
            norm = torch.sum(torch.abs(x), self.dim)
            norm[norm <= 0] = norm[norm <= 0] - 1e-6
            norm[norm >= 0] = norm[norm >= 0] + 1e-6
            norm = torch.unsqueeze(norm, self.dim)
            size = np.ones(x.dim(), dtype="int")
            size[self.dim] = x.size()[self.dim]
            norm = norm.repeat(*size)
            x = torch.div(x, norm)
        return x


class MyLoss2(Module):
    def __init__(self, thresh=1, alpha=2):
        super(MyLoss2, self).__init__()
        self.thresh = thresh
        self.alpha = alpha

    def forward(self, input1, input2):
        result = MyLoss2Function.apply(input1, input2, self.thresh, self.alpha)
        return result


class MyLoss(Module):
    def __init__(self, upper_thresh=5, lower_thresh=1):
        super(MyLoss, self).__init__()
        self.upper_thresh = 5
        self.lower_thresh = 1

    def forward(self, input1, input2):
        result = MyLossFunction.apply(
            input1, input2, self.upper_thresh, self.lower_thresh
        )
        return result


class NLFMax(Module):
    def __init__(self):
        super(NLFMax, self).__init__()

    def forward(self, input, g0, g1, g2, g3):
        result0 = NlfDownFunction.apply(input, g0)
        result1 = NlfUpFunction.apply(input, g1)
        result2 = NlfRightFunction.apply(input, g2)
        result3 = NlfLeftFunction.apply(input, g3)
        return torch.max(torch.max(torch.max(result0, result1), result2), result3)


class NLFMean(Module):
    def __init__(self):
        super(NLFMean, self).__init__()

    def forward(self, input, g0, g1, g2, g3):
        result0 = NlfDownFunction.apply(input, g0)
        result1 = NlfUpFunction.apply(input, g1)
        result2 = NlfRightFunction.apply(input, g2)
        result3 = NlfLeftFunction.apply(input, g3)
        #        result1 = NlfUpFunction()(input, g1)
        #        result2 = NlfRightFunction()(input, g2)
        #        result3 = NlfLeftFunction()(input, g3)
        #        return torch.add(torch.add(torch.add(result0, result1), result2), result3)
        return (result0 + result1 + result2 + result3) * 0.25


class NLFIter(Module):
    def __init__(self):
        super(NLFIter, self).__init__()

    def forward(self, input, g0, g1, g2, g3):
        result = NlfDownFunction.apply(input, g0)
        result = NlfUpFunction.apply(result, g1)
        result = NlfRightFunction.apply(result, g2)
        result = NlfLeftFunction.apply(result, g3)
        return result


class NLF(Module):
    def __init__(self):
        super(NLF, self).__init__()

    def forward(self, input, g0, g1, g2, g3):
        result = NlfFunction.apply(input, g0, g1, g2, g3)
        return result


class SGA(Module):
    def __init__(self):
        super(SGA, self).__init__()

    def forward(self, input, g0, g1, g2, g3):
        result = SgaFunction.apply(input, g0, g1, g2, g3)
        return result


class LGA3D3(Module):
    def __init__(self, radius=2):
        super(LGA3D3, self).__init__()
        self.radius = radius

    def forward(self, input1, input2):
        result = Lga3d3Function.apply(input1, input2, self.radius)
        return result


class LGA3D2(Module):
    def __init__(self, radius=2):
        super(LGA3D2, self).__init__()
        self.radius = radius

    def forward(self, input1, input2):
        result = Lga3d2Function.apply(input1, input2, self.radius)
        return result


class LGA3D(Module):
    def __init__(self, radius=2):
        super(LGA3D, self).__init__()
        self.radius = radius

    def forward(self, input1, input2):
        result = Lga3dFunction.apply(input1, input2, self.radius)
        return result


class LGA3(Module):
    def __init__(self, radius=2):
        super(LGA3, self).__init__()
        self.radius = radius

    def forward(self, input1, input2):
        result = Lga3Function.apply(input1, input2, self.radius)
        return result


class LGA2(Module):
    def __init__(self, radius=2):
        super(LGA2, self).__init__()
        self.radius = radius

    def forward(self, input1, input2):
        result = Lga2Function.apply(input1, input2, self.radius)
        return result


class LGA(Module):
    def __init__(self, radius=2):
        super(LGA, self).__init__()
        self.radius = radius

    def forward(self, input1, input2):
        result = LgaFunction.apply(input1, input2, self.radius)
        return result


class GetCostVolume(Module):
    def __init__(self, maxdisp):
        super(GetCostVolume, self).__init__()
        self.maxdisp = maxdisp + 1

    def forward(self, x, y):
        assert x.is_contiguous() == True
        with torch.cuda.device_of(x):
            num, channels, height, width = x.size()
            cost = (
                x.new().resize_(num, channels * 2, self.maxdisp, height, width).zero_()
            )
            #            cost = Variable(torch.FloatTensor(x.size()[0], x.size()[1]*2, self.maxdisp,  x.size()[2],  x.size()[3]).zero_(), volatile= not self.training).cuda()
            for i in range(self.maxdisp):
                if i > 0:
                    cost[:, : x.size()[1], i, :, i:] = x[:, :, :, i:]
                    cost[:, x.size()[1] :, i, :, i:] = y[:, :, :, :-i]
                else:
                    cost[:, : x.size()[1], i, :, :] = x
                    cost[:, x.size()[1] :, i, :, :] = y

            cost = cost.contiguous()
        return cost


class DisparityRegression(Module):
    def __init__(self, maxdisp):
        super(DisparityRegression, self).__init__()
        self.maxdisp = maxdisp + 1

    #        self.disp = Variable(torch.Tensor(np.reshape(np.array(range(self.maxdisp)),[1,self.maxdisp,1,1])).cuda(), requires_grad=False)

    def forward(self, x):
        assert x.is_contiguous() == True
        with torch.cuda.device_of(x):
            disp = Variable(
                torch.Tensor(
                    np.reshape(np.array(range(self.maxdisp)), [1, self.maxdisp, 1, 1])
                ).cuda(),
                requires_grad=False,
            )
            disp = disp.repeat(x.size()[0], 1, x.size()[2], x.size()[3])
            out = torch.sum(x * disp, 1)
        return out



from argparse import ArgumentParser, Namespace

import torch
import torch.nn.functional as F

from ptlflow.utils.registry import register_model
from .Networks.BOFNet.update import GMAUpdateBlock
from .Networks.encoders import twins_svt_large
from .Networks.BOFNet.cnn import BasicEncoder
from .Networks.BOFNet.corr import CorrBlock, AlternateCorrBlock
from .utils import coords_grid
from .Networks.BOFNet.gma import Attention
from .Networks.BOFNet.sk import SKUpdateBlock6_Deep_nopoolres_AllDecoder
from .Networks.BOFNet.sk2 import SKUpdateBlock6_Deep_nopoolres_AllDecoder2
from ..base_model.base_model import BaseModel


class VideoFlowBOF(BaseModel):
    pretrained_checkpoints = {
        "things_288960": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/videoflow_bof-things_288960noise-d581490a.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/videoflow_bof-sintel-c2010097.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/videoflow_bof-kitti-fa9af79c.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        cnet: str = "twins",
        fnet: str = "twins",
        gma: str = "GMA-SK2",
        pretrain: bool = True,
        corr_fn: str = "default",
        decoder_depth: int = 32,
        cost_heads_num: int = 1,
        **kwargs,
    ) -> None:
        super().__init__(loss_fn=None, output_stride=8, **kwargs)

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.cnet = cnet
        self.fnet = fnet
        self.gma = gma
        self.pretrain = pretrain
        self.corr_fn = corr_fn
        self.decoder_depth = decoder_depth

        self.hidden_dim = 128
        self.context_dim = 128

        # feature network, context network, and update block
        if cnet == "twins":
            print("[Using twins as context encoder]")
            self.cnet = twins_svt_large(pretrained=self.pretrain)
        elif cnet == "basicencoder":
            print("[Using basicencoder as context encoder]")
            self.cnet = BasicEncoder(output_dim=256, norm_fn="instance")

        if fnet == "twins":
            print("[Using twins as feature encoder]")
            self.fnet = twins_svt_large(pretrained=self.pretrain)
        elif fnet == "basicencoder":
            print("[Using basicencoder as feature encoder]")
            self.fnet = BasicEncoder(output_dim=256, norm_fn="instance")

        if self.gma == "GMA":
            print("[Using GMA]")
            self.update_block = GMAUpdateBlock(
                corr_levels=corr_levels, corr_radius=corr_radius, hidden_dim=128
            )
        elif self.gma == "GMA-SK":
            print("[Using GMA-SK]")
            self.cost_heads_num = 1
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder(
                cost_heads_num=cost_heads_num, hidden_dim=128
            )
        elif self.gma == "GMA-SK2":
            print("[Using GMA-SK2]")
            self.cost_heads_num = 1
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder2(
                corr_radius=corr_radius,
                corr_levels=corr_levels,
                cost_heads_num=cost_heads_num,
                hidden_dim=128,
            )

        print("[Using corr_fn {}]".format(self.corr_fn))

        self.att = Attention(dim=128, heads=1, max_pos_size=160, dim_head=128)

        self.has_showed_warning = False

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)
        coords1 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images = self._check_input_shape(inputs["images"])

        images, image_resizer = self.preprocess_images(
            images,
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        B, N, _, H, W = images.shape

        hdim = self.hidden_dim
        cdim = self.context_dim

        fmaps = self.fnet(images.reshape(B * N, 3, H, W)).reshape(
            B, N, -1, H // 8, W // 8
        )
        fmap1 = fmaps[:, 0, ...]
        fmap2 = fmaps[:, 1, ...]
        fmap3 = fmaps[:, 2, ...]

        if self.corr_fn == "efficient":
            corr_fn_21 = AlternateCorrBlock(fmap2, fmap1, radius=self.corr_radius)
            corr_fn_23 = AlternateCorrBlock(fmap2, fmap3, radius=self.corr_radius)
        else:
            corr_fn_21 = CorrBlock(
                fmap2,
                fmap1,
                num_levels=self.corr_levels,
                radius=self.corr_radius,
            )
            corr_fn_23 = CorrBlock(
                fmap2,
                fmap3,
                num_levels=self.corr_levels,
                radius=self.corr_radius,
            )

        cnet = self.cnet(images[:, 1, ...])
        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)
        attention = self.att(inp)

        coords0_21, coords1_21 = self.initialize_flow(images[:, 0, ...])
        coords0_23, coords1_23 = self.initialize_flow(images[:, 0, ...])

        flow_predictions = []
        for itr in range(self.decoder_depth):
            coords1_21 = coords1_21.detach()
            coords1_23 = coords1_23.detach()

            corr21 = corr_fn_21(coords1_21)
            corr23 = corr_fn_23(coords1_23)
            corr = torch.cat([corr23, corr21], dim=1)

            flow21 = coords1_21 - coords0_21
            flow23 = coords1_23 - coords0_23
            flow = torch.cat([flow23, flow21], dim=1)

            net, up_mask, delta_flow = self.update_block(
                net, inp, corr, flow, attention
            )

            up_mask_21, up_mask_23 = torch.split(up_mask, [64 * 9, 64 * 9], dim=1)

            coords1_23 = coords1_23 + delta_flow[:, 0:2, ...]
            coords1_21 = coords1_21 + delta_flow[:, 2:4, ...]

            # upsample predictions
            flow_up_23 = self.upsample_flow(coords1_23 - coords0_23, up_mask_23)
            flow_up_21 = self.upsample_flow(coords1_21 - coords0_21, up_mask_21)
            flow_up_23 = self.postprocess_predictions(
                flow_up_23, image_resizer, is_flow=True
            )
            flow_up_21 = self.postprocess_predictions(
                flow_up_21, image_resizer, is_flow=True
            )

            flow_predictions.append(torch.stack([flow_up_23, flow_up_21], dim=1))

        if self.training:
            outputs = {
                "flows": flow_up_23[:, None],
                "flows_bw": flow_up_21[:, None],
                "flow_preds": flow_predictions,
            }
        else:
            outputs = {
                "flows": flow_up_23[:, None],
                "flow_small": coords1_23 - coords0_23,
                "flow_bw_small": coords1_21 - coords0_21,
            }

        return outputs

    def _check_input_shape(self, images):
        assert (
            images.shape[1] <= 3
        ), f"videoflow_bof requires inputs of 3 frames. The current input has too many frames (found {images.shape[1]}), decrease it to 3 to run."
        if images.shape[1] == 2:
            if not self.has_showed_warning:
                print(
                    "Warning: videoflow_bof requires inputs of 3 frames, but the current input has only 2. "
                    "The first frame will be replicated to run, which may decrease the prediction accuracy. "
                    "If using validate.py or test.py, add the arguments seqlen_3-seqpos_middle to the [val/test]_dataset arg."
                )
                self.has_showed_warning = True
            images = torch.cat([images[:, :1], images], 1)
        return images


@register_model
class videoflow_bof(VideoFlowBOF):
    pass



import math

import torch
import torch.nn.functional as F
from .utils import bilinear_sampler


class LearnableCorrBlock(torch.nn.Module):
    def __init__(self, dim, num_levels=4, radius=4):
        super(LearnableCorrBlock, self).__init__()
        self.num_levels = num_levels
        self.radius = radius
        self.dim = dim

        self.raw_P = torch.nn.Parameter(torch.eye(self.dim), requires_grad=True)
        self.raw_D = torch.nn.Parameter(torch.zeros(self.dim), requires_grad=True)
        self.register_buffer("eye", torch.eye(self.dim))

    def compute_cost_volume(self, fmap1, fmap2):
        # get matrix W
        self.raw_P_upper = torch.triu(self.raw_P)
        self.skew_P = (self.raw_P_upper - self.raw_P_upper.t()) / 2
        # Cayley representation, P is in Special Orthogonal Group SO(n)
        self.P = torch.matmul(
            (self.eye + self.skew_P), torch.inverse(self.eye - self.skew_P)
        )

        # obtain the diagonal matrix with positive elements
        self.trans_D = torch.atan(self.raw_D) * 2 / math.pi
        self.D = torch.diag((1 + self.trans_D) / (1 - self.trans_D))
        self.W = torch.matmul(torch.matmul(self.P.t(), self.D), self.P)

        batch, dim, ht, wd = fmap1.shape
        fmap1 = fmap1.view(batch, dim, ht * wd)
        fmap2 = fmap2.view(batch, dim, ht * wd)

        corr = torch.matmul(torch.tensordot(fmap1, self.W, dims=[[1], [0]]), fmap2)
        corr = corr.view(batch, ht, wd, 1, ht, wd)
        corr = corr / torch.sqrt(torch.tensor(dim).float())

        batch, h1, w1, dim, h2, w2 = corr.shape
        corr = corr.view(batch * h1 * w1, dim, h2, w2)

        corr_pyramid = []
        corr_pyramid.append(corr)
        for i in range(self.num_levels):
            if min(corr.shape[2:4]) > 2 * self.radius + 1:
                corr = F.avg_pool2d(corr, 2, stride=2)
            corr_pyramid.append(corr)
        return corr_pyramid

    def forward(self, corr_pyramid, coords):
        r = self.radius
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            corr = corr_pyramid[i]
            dx = torch.linspace(-r, r, 2 * r + 1)
            dy = torch.linspace(-r, r, 2 * r + 1)
            delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
                coords.device
            )

            centroid_lvl = coords.reshape(batch * h1 * w1, 1, 1, 2) / 2**i
            delta_lvl = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
            coords_lvl = centroid_lvl + delta_lvl

            corr = bilinear_sampler(corr, coords_lvl)
            corr = corr.view(batch, h1, w1, -1)
            out_pyramid.append(corr)

        out = torch.cat(out_pyramid, dim=-1)
        return out.permute(0, 3, 1, 2)


# class CorrBlock:
#     def __init__(self, fmap1, fmap2, num_levels=4, radius=4):
#         self.num_levels = num_levels
#         self.radius = radius
#         self.corr_pyramid = []

#         # all pairs correlation
#         corr = CorrBlock.corr(fmap1, fmap2)
#         corr = corr.float()

#         batch, h1, w1, dim, h2, w2 = corr.shape
#         corr = corr.reshape(batch*h1*w1, dim, h2, w2)

#         self.corr_pyramid.append(corr)
#         for i in range(self.num_levels-1):
#             if min(corr.shape[2:4]) > 2*radius+1:
#                 corr = F.avg_pool2d(corr, 2, stride=2)
#             self.corr_pyramid.append(corr)

#     def __call__(self, coords):
#         r = self.radius
#         coords = coords.permute(0, 2, 3, 1)
#         batch, h1, w1, _ = coords.shape

#         out_pyramid = []
#         for i in range(self.num_levels):
#             corr = self.corr_pyramid[i]
#             dx = torch.linspace(-r, r, 2*r+1)
#             dy = torch.linspace(-r, r, 2*r+1)
#             delta = torch.stack(torch.meshgrid(dy, dx, indexing='ij'), axis=-1).to(coords.device)

#             centroid_lvl = coords.reshape(batch*h1*w1, 1, 1, 2) / 2**i
#             delta_lvl = delta.view(1, 2*r+1, 2*r+1, 2)
#             coords_lvl = centroid_lvl + delta_lvl

#             corr = bilinear_sampler(corr, coords_lvl)
#             corr = corr.view(batch, h1, w1, -1)
#             out_pyramid.append(corr)

#         out = torch.cat(out_pyramid, dim=-1)
#         return out.permute(0, 3, 1, 2).contiguous().float()

#     @staticmethod
#     def corr(fmap1, fmap2):
#         batch, dim, ht, wd = fmap1.shape
#         fmap1 = fmap1.view(batch, dim, ht*wd)
#         fmap2 = fmap2.view(batch, dim, ht*wd)

#         corr = torch.matmul(fmap1.transpose(1,2), fmap2)
#         corr = corr.view(batch, ht, wd, 1, ht, wd)
#         return corr  / torch.sqrt(torch.tensor(dim).float())



from ptlflow.models.base_model.base_model import BaseModel
import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from ptlflow.utils.utils import forward_interpolate_batch
from .extractor import BasicEncoder, BasicEncoderQuarter
from .update import BasicUpdateBlock, BasicUpdateBlockQuarter
from .utils import (
    coords_grid,
    coords_grid_y_first,
    upflow4,
    compute_interpolation_weights,
)
from .knn import knn_faiss_raw
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


def compute_sparse_corr(fmap1, fmap2, k=32):
    """
    Compute a cost volume containing the k-largest hypotheses for each pixel.
    Output: corr_mink
    """
    B, C, H1, W1 = fmap1.shape
    H2, W2 = fmap2.shape[2:]
    N = H1 * W1

    fmap1, fmap2 = fmap1.view(B, C, -1), fmap2.view(B, C, -1)

    with torch.no_grad():
        _, indices = knn_faiss_raw(fmap1, fmap2, k)  # [B, k, H1*W1]

        indices_coord = indices.unsqueeze(1).expand(-1, 2, -1, -1)  # [B, 2, k, H1*W1]
        coords0 = (
            coords_grid_y_first(B, H2, W2)
            .view(B, 2, 1, -1)
            .expand(-1, -1, k, -1)
            .to(fmap1.device)
        )  # [B, 2, k, H1*W1]
        coords1 = coords0.gather(3, indices_coord)  # [B, 2, k, H1*W1]
        coords1 = coords1 - coords0

        # Append batch index
        batch_index = (
            torch.arange(B).view(B, 1, 1, 1).expand(-1, -1, k, N).type_as(coords1)
        )

    # Gather by indices from map2 and compute correlation volume
    fmap2 = fmap2.gather(2, indices.view(B, 1, -1).expand(-1, C, -1)).view(B, C, k, N)
    corr_sp = torch.einsum("bcn,bckn->bkn", fmap1, fmap2).contiguous() / torch.sqrt(
        torch.tensor(C).float()
    )  # [B, k, H1*W1]

    return corr_sp, coords0, coords1, batch_index  # coords: [B, 2, k, H1*W1]


class FlowHead(nn.Module):
    def __init__(self, input_dim=256, batch_norm=True):
        super().__init__()
        if batch_norm:
            self.flowpredictor = nn.Sequential(
                nn.Conv2d(input_dim, 128, 3, padding=1),
                nn.BatchNorm2d(128),
                nn.ReLU(inplace=True),
                nn.Conv2d(128, 64, 3, padding=1),
                nn.BatchNorm2d(64),
                nn.ReLU(inplace=True),
                nn.Conv2d(64, 2, 3, padding=1),
            )
        else:
            self.flowpredictor = nn.Sequential(
                nn.Conv2d(input_dim, 128, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(128, 64, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(64, 2, 3, padding=1),
            )

    def forward(self, x):
        return self.flowpredictor(x)


class SCVBase(BaseModel):
    def __init__(
        self,
        num_k: int = 32,
        gamma: float = 0.8,
        max_flow: float = 400.0,
        iters: int = 32,
        **kwargs,
    ) -> None:
        super().__init__(
            loss_fn=SequenceLoss(gamma=gamma, max_flow=max_flow),
            output_stride=8,
            **kwargs,
        )

        self.num_k = num_k
        self.iters = iters

        try:
            import torch_scatter
        except ImportError:
            raise ImportError(
                "ERROR: torch_scatter not found."
                " SCV requires torch_scatter library to run."
                " Check instructions at: https://github.com/rusty1s/pytorch_scatter"
            )
        try:
            import faiss
        except ImportError:
            raise ImportError(
                "ERROR: faiss not found."
                " CSV requires faiss library to run."
                " Install with pip install faiss-gpu"
            )


class SCVQuarter(SCVBase):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scv-quarter-chairs-4726627e.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scv-quarter-kitti-e86c7953.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scv-quarter-sintel-2d9b4a05.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scv-quarter-things-0dac9b66.ckpt",
    }

    def __init__(
        self,
        num_k: int = 32,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        **kwargs,
    ) -> None:
        super().__init__(
            num_k=num_k, gamma=gamma, max_flow=max_flow, iters=iters, **kwargs
        )

        # feature network, context network, and update block
        self.fnet = BasicEncoderQuarter(
            output_dim=256, norm_fn="instance", dropout=False
        )
        self.cnet = BasicEncoderQuarter(output_dim=256, norm_fn="batch", dropout=False)

        # correlation volume encoder
        self.update_block = BasicUpdateBlockQuarter(hidden_dim=128, input_dim=405)

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 4, W // 4).to(img.device)
        coords1 = coords_grid(N, H // 4, W // 4).to(img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow_quarter(self, flow, mask):
        """Upsample flow field [H/4, W/4, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 4, 4, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(4 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 4 * H, 4 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        # run the feature and context network
        fmap1, fmap2 = self.fnet([image1, image2])
        cnet = self.cnet(image1)
        net, inp = torch.split(cnet, [128, 128], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)

        fmap1 = fmap1.float()
        fmap2 = fmap2.float()

        B, _, H1, W1 = fmap1.shape

        # GRU
        coords0, coords1 = self.initialize_flow(image1)

        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            forward_flow = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])
            coords1 = coords1 + forward_flow

        # Generate sparse cost volume for GRU
        corr_val, coords0_cv, coords1_cv, batch_index_cv = compute_sparse_corr(
            fmap1, fmap2, k=self.num_k
        )

        delta_flow = torch.zeros_like(coords0)

        flow_predictions = []

        search_range = 4
        corr_val = corr_val.repeat(1, 4, 1)

        for itr in range(self.iters):
            with torch.no_grad():
                # need to switch order of delta_flow, also note the minus sign
                coords1_cv = coords1_cv - delta_flow[:, [1, 0], :, :].view(
                    B, 2, 1, -1
                )  # [B, 2, k, H1*W1]

                mask_pyramid = []
                weights_pyramid = []
                coords_sparse_pyramid = []

                # Create multi-scale displacements
                for i in range(5):
                    coords1_sp = coords1_cv * 0.5**i
                    weights, coords1_sp = compute_interpolation_weights(coords1_sp)
                    mask = (coords1_sp[:, 0].abs() <= search_range) & (
                        coords1_sp[:, 1].abs() <= search_range
                    )
                    batch_ind = batch_index_cv.permute(0, 2, 3, 1).repeat(1, 4, 1, 1)[
                        mask
                    ]
                    coords0_sp = coords0_cv.permute(0, 2, 3, 1).repeat(1, 4, 1, 1)[mask]
                    coords1_sp = coords1_sp.permute(0, 2, 3, 1)[mask]

                    coords1_sp = coords1_sp + search_range
                    coords_sp = torch.cat([batch_ind, coords0_sp, coords1_sp], dim=1)
                    coords_sparse_pyramid.append(coords_sp)

                    mask_pyramid.append(mask)
                    weights_pyramid.append(weights)

            corr_val_pyramid = []
            for mask, weights in zip(mask_pyramid, weights_pyramid):
                corr_masked = (weights * corr_val)[mask].unsqueeze(1)
                corr_val_pyramid.append(corr_masked)

            sparse_tensor_pyramid = [
                torch.sparse.FloatTensor(
                    coords_sp.t().long(),
                    corr_resample,
                    torch.Size([B, H1, W1, 9, 9, 1]),
                ).coalesce()
                for coords_sp, corr_resample in zip(
                    coords_sparse_pyramid, corr_val_pyramid
                )
            ]

            corr = torch.cat(
                [sp.to_dense().view(B, H1, W1, -1) for sp in sparse_tensor_pyramid],
                dim=3,
            ).permute(0, 3, 1, 2)

            coords1 = coords1.detach()

            flow = coords1 - coords0

            # GRU Update
            # 4D net map to 2D dense vector
            net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow4(coords1 - coords0)
            else:
                flow_up = self.upsample_flow_quarter(coords1 - coords0, up_mask)

            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": coords1 - coords0}

        return outputs


class SCVEighth(SCVBase):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scv-eighth-chairs-8ba57294.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scv-eighth-things-9c893323.ckpt",
    }

    def __init__(
        self,
        num_k: int = 32,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        **kwargs,
    ) -> None:
        super().__init__(
            num_k=num_k, gamma=gamma, max_flow=max_flow, iters=iters, **kwargs
        )

        # feature network, context network, and update block
        self.fnet = BasicEncoder(output_dim=256, norm_fn="instance", dropout=False)
        self.cnet = BasicEncoder(output_dim=256, norm_fn="batch", dropout=False)

        # correlation volume encoder
        self.update_block = BasicUpdateBlock(hidden_dim=128, input_dim=405)

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8).to(img.device)
        coords1 = coords_grid(N, H // 8, W // 8).to(img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        # run the feature and context network
        fmap1, fmap2 = self.fnet([image1, image2])
        cnet = self.cnet(image1)
        net, inp = torch.split(cnet, [128, 128], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)

        fmap1 = fmap1.float()
        fmap2 = fmap2.float()

        B, _, H1, W1 = fmap1.shape

        # GRU
        coords0, coords1 = self.initialize_flow(image1)

        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            forward_flow = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])
            coords1 = coords1 + forward_flow

        # Generate sparse cost volume for GRU
        corr_val, coords0_cv, coords1_cv, batch_index_cv = compute_sparse_corr(
            fmap1, fmap2, k=self.num_k
        )

        delta_flow = torch.zeros_like(coords0)

        flow_predictions = []

        search_range = 4
        corr_val = corr_val.repeat(1, 4, 1)

        for itr in range(self.iters):
            with torch.no_grad():
                # need to switch order of delta_flow, also note the minus sign
                coords1_cv = coords1_cv - delta_flow[:, [1, 0], :, :].view(
                    B, 2, 1, -1
                )  # [B, 2, k, H1*W1]

                mask_pyramid = []
                weights_pyramid = []
                coords_sparse_pyramid = []

                # Create multi-scale displacements
                for i in range(5):
                    coords1_sp = coords1_cv * 0.5**i
                    weights, coords1_sp = compute_interpolation_weights(coords1_sp)
                    mask = (coords1_sp[:, 0].abs() <= search_range) & (
                        coords1_sp[:, 1].abs() <= search_range
                    )
                    batch_ind = batch_index_cv.permute(0, 2, 3, 1).repeat(1, 4, 1, 1)[
                        mask
                    ]
                    coords0_sp = coords0_cv.permute(0, 2, 3, 1).repeat(1, 4, 1, 1)[mask]
                    coords1_sp = coords1_sp.permute(0, 2, 3, 1)[mask]

                    coords1_sp = coords1_sp + search_range
                    coords_sp = torch.cat([batch_ind, coords0_sp, coords1_sp], dim=1)
                    coords_sparse_pyramid.append(coords_sp)

                    mask_pyramid.append(mask)
                    weights_pyramid.append(weights)

            corr_val_pyramid = []
            for mask, weights in zip(mask_pyramid, weights_pyramid):
                corr_masked = (weights * corr_val)[mask].unsqueeze(1)
                corr_val_pyramid.append(corr_masked)

            sparse_tensor_pyramid = [
                torch.sparse.FloatTensor(
                    coords_sp.t().long(),
                    corr_resample,
                    torch.Size([B, H1, W1, 9, 9, 1]),
                ).coalesce()
                for coords_sp, corr_resample in zip(
                    coords_sparse_pyramid, corr_val_pyramid
                )
            ]

            corr = torch.cat(
                [sp.to_dense().view(B, H1, W1, -1) for sp in sparse_tensor_pyramid],
                dim=3,
            ).permute(0, 3, 1, 2)

            coords1 = coords1.detach()

            flow = coords1 - coords0

            # 4D net map to 2D dense vector
            net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow4(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)

            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": coords1 - coords0}

        return outputs


@register_model
@trainable
class scv4(SCVQuarter):
    pass


@register_model
@trainable
class scv8(SCVEighth):
    pass
